/*-
 * Copyright (C) 2002-2003 Juergen Keil.
 * Copyright (C) 2002-2003 TooLs GmbH.
 * All rights reserved.
 *
 * Copyright (c) 2001 The NetBSD Foundation, Inc.
 * All rights reserved.
 *
 * This code is derived from software contributed to The NetBSD Foundation
 * by Yannick Montulet.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * A Solaris audio driver for CreativeLabs PCI sound cards based on
 * the EMU10Kx chipset.
 */

#if	_KERNEL

#include <sys/sunddi.h>
#include <sys/ksynch.h>
#include <sys/note.h>
#include <sys/mixer.h>
#include <sys/audio.h>
#include <sys/audio/audio_helper.h>
#include "emureg.h"

/*
 * Emu10k1 hardware limits
 */

#define	EMU_DMA_ALIGN	4096	/* alignment of PTE pages in physical memory */
#define	EMU_DMAMEM_NSEG	1
#define	EMU_PTESIZE	4096	/* size in bytes of an emu10k memory page */


#ifndef	EMU_32BIT_PTE
# ifdef sparc
#  define EMU_32BIT_PTE 1	/* Must enable 32-bit PTEs for SPARC DVMA */
# else
#  undef  EMU_32BIT_PTE		/* On x86, it's OK to use 31-bit phys addrs */
# endif
#endif

#if	EMU_32BIT_PTE

/* 
 * 32-Bit PTE format (might be supported with the Audigy only)
 */
#define	EMU_MAXPTE	4096	/* total number of entries in the page table */
#define	EMU_PTEMASK	0xfffff000
#define	EMU_PTIMASK	0x00000fff

#define	EMU_MK_PTE(a,i)	(((a) & EMU_PTEMASK) | ((i) & EMU_PTIMASK))
#define	EMU_PTE_ADR(pe)	((pe) & EMU_PTEMASK)
#define	EMU_PTE_IDX(pe)	((pe) & EMU_PTIMASK)
#define	EMU_MK_CHMAP(a,i) (((a) & EMU_PTEMASK) | ((i) & EMU_PTIMASK))

#else	/* !EMU_32BIT_PTE */

/* 
 * 31-Bit PTE format (should work on all EMU10Ks)
 */
#define	EMU_MAXPTE	8192	/* total number of entries in the page table */
#define	EMU_PTEMASK	0xffffe000
#define	EMU_PTIMASK	0x00001fff

#define	EMU_MK_PTE(a,i)	((((a)<<1) & EMU_PTEMASK) | ((i) & EMU_PTIMASK))
#define	EMU_PTE_ADR(pe)	(((pe) & EMU_PTEMASK) >> 1)
#define	EMU_PTE_IDX(pe)	( (pe) & EMU_PTIMASK)
#define	EMU_MK_CHMAP(a,i) ((((a)<<1) & EMU_PTEMASK) | ((i) & EMU_PTIMASK))

#endif	/* !EMU_32BIT_PTE */

#define EMU_NUMCHAN	64	/* number of independant playback channels */
#define EMU_NUMRECSRCS	3	/* number of possible record sources */

#define	SAMPLEBUF_INTR_CHUNK	(EMU_PTESIZE/2)	/* one intr every 2 KByte */
#define	SAMPLEBUF_SIZE	(12 * SAMPLEBUF_INTR_CHUNK)


/*
 * dma memory managment
 */

struct dmamem {
    ddi_dma_handle_t	dma_handle;
    uint8_t		dma_bound;
    ddi_acc_handle_t	acc_handle;
    caddr_t	       	kaddr;		/* kernel vitual address */
    uint32_t		paddr;		/* physical address 0..0x7fffffff */
    size_t		len;		/* allocated length */
    size_t		real_len;	/* total len of aligned dma mapping */
    uint_t		ccount;
    ddi_dma_cookie_t	cookie;
};


/*
 * Emu10k1 memory managment
 */

typedef struct audioemu_mem {
    struct audioemu_mem *next, **prev;
    struct dmamem      *dmamem;
    uint16_t		ptbidx;
#define	EMU_RMEM	0xFFFF		/* recording memory */
} audioemu_mem_t;

/*
 * Emu10k1 play channel params
 */

struct audioemu_chanparms_fxsend {
    struct {
	uint8_t		level, dest;
    } a, b, c, d, e, f, g, h;
};

struct audioemu_chanparms_pitch {
    uint16_t		intial;		/* 4 bits of octave, 12 bits of
					 * fractional octave */
    uint16_t		current;	/* 0x4000 == unity pitch shift */
    uint16_t		target;		/* 0x4000 == unity pitch shift */
    uint8_t		envelope_amount;/* Signed 2's complement, +/-
					 * one octave peak extremes */
};

struct audioemu_chanparms_envelope {
    uint16_t		current_state;	/* 0x8000-n == 666*n usec delay */
    uint8_t		hold_time;	/* 127-n == n*(volume ? 88.2 :
					 * 42)msec */
    uint8_t		attack_time;	/* 0 = infinite, 1 = (volume ? 11 :
					 * 10.9) msec, 0x7f = 5.5msec */
    uint8_t		sustain_level;	/* 127 = full, 0 = off, 0.75dB
					 * increments */
    uint8_t		decay_time;	/* 0 = 43.7msec, 1 = 21.8msec, 0x7f =
					 * 22msec */
};

struct audioemu_chanparms_volume {
    uint16_t		current, target;
    struct audioemu_chanparms_envelope envelope;
};

struct audioemu_chanparms_filter {
    uint16_t		initial_cutoff_frequency;
    /*
     * 6 most  significant bits are semitones, 2 least significant bits
     * are fractions
     */
    uint16_t		current_cutoff_frequency;
    uint16_t		target_cutoff_frequency;
    uint8_t		lowpass_resonance_height;
    uint8_t		interpolation_ROM;	/* 1 = full band, 7 = low
						 * pass */
    uint8_t		envelope_amount;	/* Signed 2's complement, +/-
						 * six octaves peak extremes */
    uint8_t		LFO_modulation_depth;	/* Signed 2's complement, +/-
						 * three octave extremes */
};

struct audioemu_chanparms_loop {
    uint32_t		start;		/* index in the PTB (in samples) */
    uint32_t		end;		/* index in the PTB (in samples) */
};

struct audioemu_chanparms_modulation {
    struct audioemu_chanparms_envelope envelope;
    uint16_t		LFO_state;	/* 0x8000-n = 666*n usec delay */
};

struct audioemu_chanparms_vibrato_LFO {
    uint16_t		state;		/* 0x8000-n == 666*n usec delay */
    uint8_t		modulation_depth;/* Signed 2's complement, +/-
					  * one octave extremes */
    uint8_t		vibrato_depth;	/* Signed 2's complement, +/- one
					 * octave extremes */
    uint8_t		frequency;	/* 0.039Hz steps, maximum of 9.85 Hz */
};

typedef struct audioemu_channel {
    uint8_t		num;	/* voice number */
    struct audioemu_voice *voice;
    struct audioemu_chanparms_fxsend fxsend;
    struct audioemu_chanparms_pitch pitch;
    uint16_t		initial_attenuation;	/* 0.375dB steps */
    struct audioemu_chanparms_volume volume;
    struct audioemu_chanparms_filter filter;
    struct audioemu_chanparms_loop loop;
    struct audioemu_chanparms_modulation modulation;
    struct audioemu_chanparms_vibrato_LFO vibrato_LFO;
    uint8_t		tremolo_depth;
} audioemu_channel_t;


/*
 * Voices, streams
 */

typedef enum {
    EMU_RECSRC_MIC = 0,
    EMU_RECSRC_ADC,
    EMU_RECSRC_FX,
    EMU_RECSRC_NOTSET
} audioemu_recsrc_t;


typedef struct audioemu_voice {
    struct audioemu_softstate *softstate;	/* our softstate */

    uint8_t		use;
#define	EMU_VOICE_USE_PLAY	(1 << 0)
#define	EMU_VOICE_USE_REC_MIC	(EMU_RECSRC_MIC << 1)
#define	EMU_VOICE_USE_REC_ADC	(EMU_RECSRC_ADC << 1)
#define	EMU_VOICE_USE_REC_FX	(EMU_RECSRC_FX  << 1)
#define	EMU_VOICE_USE_RECSRC(x)	(((x) >> 1) & 0x3)
    uint8_t		state;
#define EMU_VOICE_STATE_STARTED	(1 << 0)
    uint8_t		stereo;
#define	EMU_VOICE_STEREO_NOTSET	0xFF
    uint8_t		b16;
    uint32_t		sample_rate;
    union {
	audioemu_channel_t *chan[2];
	audioemu_recsrc_t source;
    } dataloc;
    audioemu_mem_t     *buffer;
    int			offs;	    /* byte offset of 1st valid sample */
    int			used;	    /* #bytes of valid sample data */
    uint16_t		blksize;    /* in samples */
    uint16_t		trigblk;    /* blk on which to trigger inth */
    uint16_t		blkmod;	    /* Modulo value to wrap trigblk */
    uint16_t		timerate;
    void	      (*inth) (void *);
    void	       *inthparam;

    struct audioemu_voice *next, **prev;
} audioemu_voice_t;


/*
 * Per device softstate
 */
typedef struct audioemu_softstate {
    dev_info_t         *dip;            /* device info node for your sound hw */
    audiohdl_t		audio_handle;	/* opaque handle to audio support module */

    audio_codec_handle_t codec_handle;

    kmutex_t            instance_mutex;
    kmutex_t            intr_mutex;
    ddi_iblock_cookie_t	intr_cookie;

    caddr_t		regs;		/* audigy I/O space registers */
    ddi_acc_handle_t	regs_handle;

    ddi_acc_handle_t	pci_conf_handle;

    struct audioemu_mem *mem;		/* list of dma mem chunks */

    struct dmamem      *ptb;
    struct dmamem      *silentpage;

    audioemu_channel_t *channel[EMU_NUMCHAN];
    audioemu_voice_t   *recsrc[EMU_NUMRECSRCS];
    kmutex_t            voices_lock;
    audioemu_voice_t   *voices;		/* list of allocated voices */

    uint8_t		timerstate;
#define	EMU_TIMER_STATE_ENABLED	1
    uint8_t		reschedule;
#define	EMU_TIMER_RESCHED_PENDING	1

    audioemu_voice_t   *pvoice;		/* playback voice */
    audioemu_voice_t   *rvoice;		/* record voice */
    audioemu_voice_t   *lvoice;

    void	       *pbuf;
    void	       *rbuf;

    char               *revstr;         /* pci revision of card as printable string */
    uint8_t             rev;            /* pci revision byte of card */
    uint8_t		audigy;		/* 0: SB Live!
					 * 1: SB Audigy
					 * 2: SB Audigy2
					 */

    uint16_t		ac97_ext_features;

    audio_info_t	audio_info;	/* driver local audio state. Note: */
    					/* we maintain only a few fields */
    					/* of the structure in the driver */

    audio_device_t	dev_info;
    ah_ad_info_t	ad_info;
} audioemu_softstate_t;


#define	TRUE	1
#define	FALSE	0

/* general success/failure return codes */
#define	AUDIOEMU_SUCCESS	0
#define	AUDIOEMU_FAILURE	(-1)


/*
 * macros to access the audigy register block
 */
#define	EMU_GET8(offs)						\
	ddi_get8(softstate->regs_handle,			\
		  (void*) ((char*)softstate->regs + (offs)))
#define	EMU_GET16(offs)						\
	ddi_get16(softstate->regs_handle,			\
		  (void*) ((char*)softstate->regs + (offs)))
#define	EMU_GET32(offs)						\
	ddi_get32(softstate->regs_handle,			\
		  (void*) ((char*)softstate->regs + (offs)))

#define	EMU_PUT8(offs, val)					\
	ddi_put8(softstate->regs_handle,			\
		  (void*) ((char*)softstate->regs + (offs)),	\
		  (val))
#define	EMU_PUT16(offs, val)					\
	ddi_put16(softstate->regs_handle,			\
		  (void*) ((char*)softstate->regs + (offs)),	\
		  (val))
#define	EMU_PUT32(offs, val)					\
	ddi_put32(softstate->regs_handle,			\
		  (void*) ((char*)softstate->regs + (offs)),	\
		  (val))


#define	EMU_REGS_SIZE	0x20	/* size of audigy register block */
#define	EMU_A2_REGS_SIZE 0x40	/* size of audigy 2 register block */



#endif  /* _KERNEL */


/* custom 'ioctl' supported by the audioemu driver */
#define	AUDIOEMU_IOC			('i'<<8)
#define	AUDIOEMU_SET_PCMOUT_VOL 	(AUDIOEMU_IOC | 0)
#define	AUDIOEMU_SET_BEEP_VOL		(AUDIOEMU_IOC | 1)
#define	AUDIOEMU_SET_LOUDNESS_ENABLE	(AUDIOEMU_IOC | 2)
#define	AUDIOEMU_SET_3D_STEREO_ENABLE	(AUDIOEMU_IOC | 3)
#define	AUDIOEMU_SET_3D_STEREO_DEPTH	(AUDIOEMU_IOC | 4)
#define	AUDIOEMU_SET_3D_STEREO_CENTER	(AUDIOEMU_IOC | 5)
#define	AUDIOEMU_SET_STEREO_ENHANCE_ENABLE	(AUDIOEMU_IOC | 6)

