/*-
 * Copyright (C) 2002 Juergen Keil.
 * Copyright (C) 2002 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * This file contains chipset specific code for the Ensoniq 1370 audio
 * controller.  This includes the Asahi Kasei AK4531 audio codec
 * interface, the fixed frequencies available only for the "FM" DAC1
 * and the PLL clock for the "VOICE" DAC2 and the ADC.
 */

#ifdef	__GNUC__
#include <stdarg.h>
#define _SYS_VARARGS_H          /* avoid including sys/varargs.h with gcc */
#else
#include <sys/varargs.h>
#endif

#include <sys/modctl.h>
#include <sys/file.h>
#include <sys/stream.h>
#include <sys/conf.h>
#include <sys/pci.h>
#include <sys/audio.h>
#include "audioio_compat.h"
#include "audio1371.h"
#include "akm4531_codec.h"

#include <sys/ddi.h>


/************************************************************/
/* Start of CODEC section.                                  */
/************************************************************/


/*
 * Waits for a while until the CWRIP or CSTAT bit is clear.
 *
 * Returns AUDIO1371_SUCCESS if the code is idle or AUDIO1371_FAILURE if
 * the codec is still busy after a maximum wait of 40msecs.
 */

static int
ens1370_wait_for_codecclear(struct audio1371_softstate *softstate,
			    uint32_t waitstat)
{
    int loop;
    uint32_t value;

    for (loop = 2000; --loop >= 0; ) {
	value = SBPCI_GET32(SBPCI_STATUS);
	if ((value & waitstat) == 0) {
	    /* codec is idle now, no write in progress */
	    return AUDIO1371_SUCCESS;
	}
	drv_usecwait(20);
    }

    return AUDIO1371_FAILURE;
}


/*
 * Write a AKM4531 codec control register.
 * returns TRUE if the write op could be submitted to the HW, and FALSE
 * otherwise.
 *
 * Note: the ENS1370 can only write to the AKM4531, the hardware does
 * not povide read access to the codec registers on the AKM4531.
 */
static int 
ens1370_codec_write(void *closure, uint8_t sub_addr, uint8_t value)
{
    struct audio1371_softstate *softstate = closure;
    uint32_t codecvalue;
    uint32_t waitstat;

    /*
     * The reset&power_down register can be written even if the AK4531
     * codec is busy; we just wait for a codec write in progress to
     * finsh.  For all other codec register accesses we wait until no
     * codec write is in progress any more and the codec is not busy.
     */
    waitstat = (sub_addr == 0x16 /*AK_RESET_POWER_DOWN*/)
	? SBPCI_1370_STAT_CWRIP
	: SBPCI_1370_STAT_CSTAT;

    if (ens1370_wait_for_codecclear(softstate, waitstat) != AUDIO1371_SUCCESS) {
	DPRINTF2(AUD1371M_CODEC,
		 "ens1370_write: codec not idle (setting R%02x to %02x)",
		 sub_addr, value);
	return FALSE;
    }

    codecvalue = (sub_addr & 0xff) << 8;
    codecvalue |= (value & 0xff);

    SBPCI_PUT16(SBPCI_1370_CODEC, codecvalue);

    return TRUE;
}


static int 
ens1370_codec_wait_idle(void *closure)
{
    struct audio1371_softstate *softstate = closure;
    int i;
    int stat;

    /*
     * wait while the codec is busy initializing.
     *
     * According to the documentation, this state can last for 516/fs
     * and with a low frequency of 5.512 kHz this would be ~90 msec.
     *
     * Since ens1370_wait_for_codecclear waits for only 40 msec, we 
     * have to retry several times to wait for the end of the busy
     * period after a codec reset / power up.
     */
    for (i = 0; i < 5; i++) {
	stat = ens1370_wait_for_codecclear(softstate, SBPCI_1370_STAT_CSTAT);
	if (stat == AUDIO1371_SUCCESS)
	    return TRUE;
    }
    DPRINTF2(AUD1371M_CODEC, "ens1370_codec_wait_idle: codec not idle");
    return FALSE;
}


/*
 * Initialize the link between audio controller and audio codec.
 */
static void
ens1370_codec_interface_reset(struct audio1371_softstate *softstate)
{
    uint32_t ctrl;
    uint32_t pclkdiv;

    /* enable the 1370 codec interface */
    ctrl = SBPCI_GET32(SBPCI_CONTROL);
    ctrl = SBPCI_1370_CTRL_CODEC | SBPCI_1370_CTRL_SERR_DIS;
    pclkdiv = SBPCI_1370_SR2PCLKDIV(8000);
    ctrl |= (pclkdiv << SBPCI_1370_CTRL_PCLKDIV_SHIFT)
			& SBPCI_1370_CTRL_PCLKDIV_MASK;
    /* ctrl |= SBPCI_1370_CTRL_44100; */
    SBPCI_PUT32(SBPCI_CONTROL, ctrl);
}


/*
 * Basic configuration of the audio codec.
 */
static audio_codec_handle_t
ens1370_codec_init(struct audio1371_softstate *softstate)
{
    static struct akm4531_codec_ops codec_ops = {
	ens1370_codec_write,
	ens1370_codec_wait_idle
    };

    return akm4531_codec_init(softstate->audio_handle, &codec_ops, softstate);
}


static int 
ens1370_configure_pm(struct audio1371_softstate *softstate, int level)
{
    /* Power management is not available on ens1370 card */
    return AUDIO1371_SUCCESS;
}


static int 
ens1370_init_sample_rate_conv(struct audio1371_softstate *softstate)
{
    /* not available on ens1370 card */
    return AUDIO1371_SUCCESS;
}


static int 
ens1370_enable_sample_rate_conv(struct audio1371_softstate *softstate,
				int dacnum,
				int enable_flag,
				int *old_enable_flag)
{
    /* not available on ens1370 card */
    return AUDIO1371_SUCCESS;
}


/*
 * Set the record ADC samplerate by configuring the PLL clock.
 * rate is in the range 4000 to 48000 (hz).
 *
 * FIXME XXX: ADC and DAC2 run from the same PLL clock, so we should
 * refuse different ADC / DAC2 rates.
 */
static void
ens1370_setADCfreq(struct audio1371_softstate *softstate, int rate)
{
    uint32_t ctrl;
    uint32_t pclkdiv;

    ctrl = SBPCI_GET32(SBPCI_CONTROL);

    if (rate < 4000)
	rate = 4000;
    else if (rate > 48000)
	rate = 48000;

    ctrl &= ~SBPCI_1370_CTRL_PCLKDIV_MASK;
    pclkdiv = SBPCI_1370_SR2PCLKDIV(rate);
    ctrl |= (pclkdiv << SBPCI_1370_CTRL_PCLKDIV_SHIFT)
	& SBPCI_1370_CTRL_PCLKDIV_MASK;

    SBPCI_PUT32(SBPCI_CONTROL, ctrl);
    /*printf("ADC rate=%d, ctrl=%x\n", rate, ctrl);*/
}


/* 
 * dacnum 1:
 * CODEC FM with fixed frequencies, 5.512kHz/11.025kHz/22.05kHz/44.1kHz
 * dacnum 2:
 * CODEC VOICE, variable PLL clock, rate is in the range 4kHz to 48kHz
 *
 * FIXME XXX: ADC and DAC2 run from the same PLL clock, so we should
 * refuse different ADC / DAC2 rates.
 */
static void
ens1370_setDACfreq(struct audio1371_softstate *softstate, int rate, int dacnum)
{
    uint32_t ctrl;
    uint32_t pclkdiv;

    ctrl = SBPCI_GET32(SBPCI_CONTROL);

    if (dacnum == 1) {
	ctrl &= ~SBPCI_1370_CTRL_WTSRSEL;
	if (rate >= 44100) 
	    ctrl |= SBPCI_1370_CTRL_44100;
	else if (rate >= 22050) 
	    ctrl |= SBPCI_1370_CTRL_22050;
	else if (rate >= 11025) 
	    ctrl |= SBPCI_1370_CTRL_11025;
	else 
	    ctrl |= SBPCI_1370_CTRL_5512;
	    
    } else {
	if (rate < 4000)
	    rate = 4000;
	else if (rate > 48000)
	    rate = 48000;

	ctrl &= ~SBPCI_1370_CTRL_PCLKDIV_MASK;
	pclkdiv = SBPCI_1370_SR2PCLKDIV(rate);
	ctrl |= (pclkdiv << SBPCI_1370_CTRL_PCLKDIV_SHIFT)
		    & SBPCI_1370_CTRL_PCLKDIV_MASK;
    }

    SBPCI_PUT32(SBPCI_CONTROL, ctrl);
    /*printf("DAC%d rate=%d, ctrl=%x\n", dacnum, rate, ctrl);*/

    /*
     * Messing with the clock rates may trigger a longer codec-is-busy
     * period. Wait until the codec is idle, so that we don't fail on 
     * the next codec write access. 
     *
     * (XXX maybe we need a longer timeout in ens1370_wait_for_codecclear()?)
     */
    ens1370_codec_wait_idle(softstate); 
}


ens_chip_ops_t ens1370_ops = {
    ens1370_codec_init,

    ens1370_codec_interface_reset,

    ens1370_configure_pm,

    ens1370_init_sample_rate_conv,
    ens1370_enable_sample_rate_conv,
    ens1370_setADCfreq,
    ens1370_setDACfreq
};
