/*-
 * Copyright (C) 2002 Juergen Keil.
 * Copyright (C) 2002 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#ifdef	__GNUC__
#include <stdarg.h>
#define _SYS_VARARGS_H          /* avoid including sys/varargs.h with gcc */
#else
#include <sys/varargs.h>
#endif

#include <sys/modctl.h>
#include <sys/file.h>
#include <sys/stream.h>
#include <sys/conf.h>
#include <sys/pci.h>
#include <sys/audio.h>
#include "audioio_compat.h"
#include "ac97.h"
#include "ac97_codec.h"

#include <sys/sunddi.h>
#include <sys/ddi.h>


#define	DPRINTF4(...)	audio_hlp_log(ac->audio_handle, 4, __VA_ARGS__)
#define	DPRINTF3(...)	audio_hlp_log(ac->audio_handle, 3, __VA_ARGS__)
#define	DPRINTF2(...)	audio_hlp_log(ac->audio_handle, 2, __VA_ARGS__)
#define	DPRINTF1(...)	audio_hlp_log(ac->audio_handle, 1, __VA_ARGS__)
#define	DPRINTF0(...)	audio_hlp_log(ac->audio_handle, 0, __VA_ARGS__)


/*
 * Cleanup routine, when a codec object is no longer used.
 */
static void
ac97_delete(struct audio_codec *base)
{
    struct ac97_codec *ac = (struct ac97_codec *)base;
    kmem_free(ac, sizeof(struct ac97_codec));
}


/*
 * helper subroutine to convert sun audio_info gain values
 * to ac97 codec playback volume levels.  ac97 volume is an
 * "attenuation",  lower values means "louder".
 *
 * we use a linear mapping of [0..maxvol] -> [codec_max..0]
 */
int
ac97_codec_play_vol(int vol, int maxvol, int codec_max)
{
    if (vol < 0 || vol > maxvol) {
	/*
	 * openwin audiotool sends illegal "balance" values,
	 * handle them gracefully.
	 *
	DPRINTF4("illegal volume %d", vol);
	 */
	if (vol < 0)
	    vol = 0;
	else if (vol > maxvol)
	    vol = maxvol;
    }

    return (maxvol - vol) * codec_max / maxvol;
}


/*
 * helper subroutine to convert sun audio_info gain values
 * to ac97 codec record gain levels.  ac97 record gain is
 * also a "gain" value, higher values means "louder".
 *
 * we use a linear mapping of [0..maxvol] -> [0..codec_max]
 */
static int
ac97_codec_record_vol(int vol, int maxvol, int codec_max)
{
    if (vol < 0 || vol > maxvol) {
	/*
	DPRINTF4("illegal volume %d", vol);
	*/
	if (vol < 0)
	    vol = 0;
	else if (vol > maxvol)
	    vol = maxvol;
    }

    return vol * codec_max / maxvol;
}


static void
ac97_configure_playback_vol(struct ac97_codec *ac,
			    audio_info_t      *audio_info)
{
    audio_prinfo_t *play = &audio_info->play;
    int voll, volr;
    uint16_t codecval;
    int lfactor, rfactor;	/* 0.0 .. 1.0 (scaled integer) */
#define	BALANCE_SCALE	(8)

    if (play->balance >= AUDIO_MID_BALANCE) {
	/* 
	 * MID_BALANCE..RIGHT_BALANCE
	 * full gain on right channel, reduced gain on left channel.
	 */
	lfactor = AUDIO_RIGHT_BALANCE - play->balance;
	lfactor = (lfactor << BALANCE_SCALE) / AUDIO_MID_BALANCE;
	rfactor = 1 << BALANCE_SCALE;
    } else {
	/* 
	 * LEFT_BALANCE..MID_BALANCE
	 * full gain on left channel, reduced gain on right channel.
	 */
	lfactor = 1 << BALANCE_SCALE;
	rfactor = play->balance;
	rfactor = (rfactor << BALANCE_SCALE) / AUDIO_MID_BALANCE;
    }
    voll = ac97_codec_play_vol((play->gain * lfactor) >> BALANCE_SCALE,
			       AUDIO_MAX_GAIN,
			       ac->ac97_mvol_max);
    volr = ac97_codec_play_vol((play->gain * rfactor) >> BALANCE_SCALE,
			       AUDIO_MAX_GAIN,
			       ac->ac97_mvol_max);

    /* ac97: master volume (speaker out jack) */
    codecval = (voll << 8) | volr;
    if (audio_info->output_muted || !(play->port & (AUDIO_SPEAKER|AUDIO_HEADPHONE)))
	codecval |= 0x8000;
    AC97_WRITE(ac, AC97_MASTERVOL, codecval);
    DPRINTF4("MASTERVOL=%x", codecval);

    /* opt: Aux out volume (line out jack) */
    codecval = (voll << 8) | volr;
    if (audio_info->output_muted || !(play->port & AUDIO_LINE_OUT))
	codecval |= 0x8000;
    AC97_WRITE(ac, AC97_AUXOUTVOL, codecval);
    DPRINTF4("AUXOUTVOL=%x", codecval);

    /* opt: MONO_OUT c*/
    codecval = ac97_codec_play_vol(play->gain,
				   AUDIO_MAX_GAIN,
				   ac->ac97_mvol_max);
    if (audio_info->output_muted)
	codecval |= 0x8000;
    AC97_WRITE(ac, AC97_MONOVOL,   codecval);


    if (ac->ad1980) {
	codecval = (voll << 8) | volr;
	if (audio_info->output_muted || !(play->port & (AUDIO_SPEAKER|AUDIO_HEADPHONE)))
	    codecval |= 0x8080;
	AC97_WRITE(ac, AC97_VOL2, codecval);
	DPRINTF4("SURRVOL=%x", codecval);
    }
}


static void
ac97_configure_record_vol(struct ac97_codec *ac,
			  audio_info_t	    *audio_info)
{
    audio_prinfo_t *record = &audio_info->record;
    int lfactor, rfactor;	/* 0.0 .. 1.0 (scaled integer) */
    int vol, voll, volr;
    uint16_t codecval;
    int record_source;

    vol = ac97_codec_play_vol(audio_info->monitor_gain, AUDIO_MAX_GAIN, 0x1f);
    /* 
     * audio_info_t is missing fields for changing the balance of
     * the monitor volume.  We have to assume AUDIO_MID_BALANCE.
     */
    voll = volr = vol;

    /*
     * the following AC97 codec registers are for *playback*
     * (monitoring of recording)
     * the ac97 mixer can output all of these (+ the PCM playback)
     * concurrently.
     */
    codecval = vol;
    if (ac->ac97_mic_20db_boost)
	codecval |= 0x40 /* 20db boost */;

    if (!(record->port & AUDIO_MICROPHONE))
	codecval |= 0x8000;	/* mute bit */
    AC97_WRITE(ac, AC97_MICVOL, codecval);
    DPRINTF4("MICVOL=%x", codecval);

    codecval = (voll << 8) | volr;
    if (!(record->port & AUDIO_LINE_IN))
	codecval |= 0x8000;	/* mute bit */
    AC97_WRITE(ac, AC97_LINEINVOL, codecval);
    DPRINTF4("LINEINVOL=%x", codecval);

    codecval = (voll << 8) | volr;
    if (!(record->port & AUDIO_CD))
	codecval |= 0x8000;	/* mute bit */
    AC97_WRITE(ac, AC97_CDVOL, codecval);
    DPRINTF4("CDVOL=%x", codecval);

    codecval = vol;
    if (!(record->port & AUDIO_AUX1_IN))
	codecval |= 0x8000;	/* mute bit */
    AC97_WRITE(ac, AC97_PHONEVOL, codecval);
    DPRINTF4("PHONEVOL=%x", codecval);

    codecval = (voll << 8) | volr;
    if (!(record->port & AUDIO_AUX2_IN))
	codecval |= 0x8000;	/* mute bit */
    AC97_WRITE(ac, AC97_VIDEOVOL, codecval);
    DPRINTF4("VIDEOVOL=%x", codecval);


    /*
     * The following AC97 codec registers are for *recording*.
     * we can choose exactly one record input source for both the
     * left and right channel.  Since audio_info cannot specify
     * independand left and right channel input source,  we always
     * have to configure both to the same record source.
     */
    vol = ac97_codec_record_vol(record->gain, AUDIO_MAX_GAIN, 0xf);

    if (record->balance >= AUDIO_MID_BALANCE) {
	/* 
	 * MID_BALANCE..RIGHT_BALANCE
	 * full gain on right channel, reduced gain on left channel.
	 */
	lfactor = AUDIO_RIGHT_BALANCE - record->balance;
	lfactor = (lfactor << BALANCE_SCALE) / AUDIO_MID_BALANCE;
	rfactor = 1 << BALANCE_SCALE;
    } else {
	/* 
	 * LEFT_BALANCE..MID_BALANCE
	 * full gain on left channel, reduced gain on right channel.
	 */
	lfactor = 1 << BALANCE_SCALE;
	rfactor = record->balance;
	rfactor = (rfactor << BALANCE_SCALE) / AUDIO_MID_BALANCE;
    }

    voll = ac97_codec_record_vol((record->gain * lfactor) >> BALANCE_SCALE,
				 AUDIO_MAX_GAIN,
				 0xf);
    volr = ac97_codec_record_vol((record->gain * rfactor) >> BALANCE_SCALE,
				 AUDIO_MAX_GAIN,
				 0xf);
    codecval = (voll << 8) | volr;
    AC97_WRITE(ac, AC97_RECGAIN,    codecval);
    DPRINTF4("RECGAIN=%x",    codecval);
    AC97_WRITE(ac, AC97_RECGAINMIC, vol);
    DPRINTF4("RECGAINMIC=%x", vol);

    record_source = 0;
    if (record->port & AUDIO_MICROPHONE)
	record_source = 0 /* AC97 record source: microphone */;
    if (record->port & AUDIO_LINE_IN)
	record_source = 4 /* AC97 record source: line-in */;
    if (record->port & AUDIO_CD)
	record_source = 1 /* AC97 record source: CD */;
    if (record->port & AUDIO_AUX1_IN)
	record_source = 7 /* AC97 record source: Phone */;
    if (record->port & AUDIO_AUX2_IN)
	record_source = 2 /* AC97 record source: Video-in */;
    if (record->port & AUDIO_CODEC_LOOPB_IN)
	record_source = 5 /* AC97 record source: stero mix */;
    AC97_WRITE(ac, AC97_RECSELCTL,
	       (record_source << 8) | record_source);
}


/*
 * First time initialization of the audio codec.
 *
 * We reset the AC97 here. This will initialize all codec register to
 * default values.
 * 
 * Some global audio codec settings are performed as well.
 */
static int
ac97_reset(struct audio_codec *base, audio_info_t *audio_info, int verbose)
{
    struct ac97_codec *ac = (struct ac97_codec *)base;
    uint16_t codecval, reset_reg, xid, pwr_stat, vid1, vid2;
    int vol;
    char *auxout, *codec_addr, *codec_rev;
    
    if (!AC97_READ(ac, AC97_POWERDOWN, &pwr_stat)) {
	cmn_err(CE_NOTE, "ac97 powerdown register read failed");
	return 0;
    }

    /* ac97 audio register reset to default values */
    if (!AC97_WRITE(ac, AC97_RESETREG, 0)) {
	cmn_err(CE_NOTE, "ac97 reset register write failed");
	return 0;
    }

    AC97_READ (ac, AC97_RESETREG, &reset_reg);
    AC97_READ (ac, AC97_XID, &xid);

    /* determine the type of the aux out port */
    if ((reset_reg & 0x0010)		/* headphone out? */
	&& AC97_READ(ac, AC97_AUXOUTVOL, &codecval)
	&& codecval == 0x8000		/* auxout vol implmented? */ ) {
	/*
	 * aux_out: implemented as headphone_out
	 * headphone amplifier can drive 22 ohm headphone
	 * volume can be controlled via auxout volume control
	 */
	ac->ac97_auxout = AUX_HP_OUT;
	auxout = "HP_OUT";
    } else if ((xid & 0x0080)		/* surround dac? */
	       && AC97_READ(ac, AC97_VOL2, &codecval)
	       && codecval == 0x8080	/* surround vol implemented? */ ) {
	/*
	 * aux_out: implemented as 4channel L&R surround
	 */
	ac->ac97_auxout = AUX_4CH_OUT;
	auxout = "4CH_OUT";
    } else {
	/*
	 * aux_out: true line level without volume control
	 */
	ac->ac97_auxout = AUX_LNLVL_OUT;
	auxout = "LNLVL_OUT";
    }

    if ((xid >> 14) & 0x3)
	codec_addr = "secondary";
    else
	codec_addr = "primary";

    switch ((xid >> 10) & 0x3) {
    case 0:
	codec_rev = "2.1 or earlier";
	ac->ac97_rev = 21;
	break;
    case 1:
	codec_rev = "2.2";
	ac->ac97_rev = 22;
	break;
    case 2:
	codec_rev = "2.3";
	ac->ac97_rev = 23;
	break;
    default:
	codec_rev = "unknown";
	ac->ac97_rev = 0;
	break;
    }

    AC97_READ (ac, AC97_VENDOR1, &vid1);
    AC97_READ (ac, AC97_VENDOR2, &vid2);

    if (verbose) {

	cmn_err(CE_CONT, 
		"?AC97: %s codec, rev. %s, "
		"vendor id1:0x%04x,id2:0x%04x,"
		"%s%s%s%s%s"
		" %d-bit DAC, %d-bit ADC, 3D-stereo mode %d,"
		" aux %s\n",
		codec_addr, codec_rev,
		vid1, vid2,
		reset_reg & 0x0001 ? " Mic PCM in," : "",
		reset_reg & 0x0004 ? " Bass/Treble," : "",
		reset_reg & 0x0008 ? " Simulated Stereo," : "",
		reset_reg & 0x0010 ? " Headphone out," : "",
		reset_reg & 0x0020 ? " Loudness," : "",
		reset_reg & 0x0040 ? 18
	    			  : reset_reg & 0x0080 ? 20 : 16,
		reset_reg & 0x0100 ? 18
				  : reset_reg & 0x0200 ? 20 : 16,
		(reset_reg & 0x7c00) >> 10,
		auxout);

	if (xid || pwr_stat)	/* any extended audio features available? */
	    cmn_err(CE_CONT, "?AC97: extid 0x%04x, powerdown status 0x%04x\n",
		    xid, pwr_stat);
    }

    /* determine if the codec implements a 5 or 6 bit volume register */
    AC97_WRITE(ac, AC97_MASTERVOL, 0x8000 | 0x3f3f);
    AC97_READ (ac, AC97_MASTERVOL, &codecval);
    DPRINTF4("MASTERVOL 0x3f3f -> %x", codecval);
    ac->ac97_mvol_max = (codecval & 0x7fff) == 0x1f1f ? 0x1f : 0x3f;

    AC97_WRITE(ac, AC97_BEEPVOL,   0x0000);

    /*
     * configure ac97 input and output ports (master volume, indiviual
     * mixer input input volumnes, the record source and the record gain)
     */
    ac97_configure_playback_vol(ac, audio_info);
    ac97_configure_record_vol(ac, audio_info);

    vol = ac->ac97_pcmvol;
    AC97_WRITE(ac, AC97_PCMVOL,    (vol << 8) | vol);

    if (ac->ac97_eapd != -1) {
	/*
	 * optional: disable/power down external amplifier. Apparently some
	 * systems have wired this backwards, and setting the disable bit
	 * enables the amplifier.
	 */
	if (ac->ac97_eapd)
	    AC97_WRITE(ac, AC97_POWERDOWN, pwr_stat |  0x8000);
	else
	    AC97_WRITE(ac, AC97_POWERDOWN, pwr_stat & ~0x8000);
    }

#if 0
    /*
     * Disable the Jack Sense pins on AD1885 codec.
     * (note sure if we need this, the user reported it started to work
     * without this change)
     */
    if (vid1 == 0x4144 && vid2 == 0x5360) {
#define	AC97_AD_JACKSENSE 0x72
#define	AD1885_JS0DIS	0x0100	/* 1: codec ignores the jack sense pin JS0 */
#define	AD1885_JS1DIS	0x0200	/* 1: codec ignores the jack sense pin JS1 */
    	AC97_READ (ac, AC97_AD_JACKSENSE, &codecval);
    	AC97_WRITE(ac, AC97_AD_JACKSENSE, codecval | AD1885_JS0DIS |
	    AD1885_JS1DIS);
    }
#endif

    /*
     * With Analog Devices AD1888/1980/1985 AC97 codecs, is seems most vendors
     * connect the surround jack to the codec's LINE_OUT signal and the
     * line out jack to the SURR_OUT signal.
     * Swap the signals in the AD1888/1980/1985 codec.
     */
    ac->ad1980 = (vid1 == 0x4144 && vid2 == 0x5368) ||
	(vid1 == 0x4144 && vid2 == 0x5370) ||
	(vid1 == 0x4144 && vid2 == 0x5375);
    if (ac->ad1980) {
#define	AC97_AD_MISC	0x76
#define	AD1980_HPSEL	0x0400
#define	AD1980_LOSEL	0x0020
    	AC97_READ (ac, AC97_AD_MISC, &codecval);
    	AC97_WRITE(ac, AC97_AD_MISC, codecval | AD1980_HPSEL | AD1980_LOSEL);
    }

    /* workaround: line-out does not work with Sigmatel STAC 9758 codec */
    if (vid1 == 0x8384 && vid2 == 0x7658) {
#define AC97_STAC9758_OUTPUT_SELECT    0x64
#define AC97_STAC9758_INPUT_SELECT     0x66
#define AC97_STAC9758_IO_MISC          0x68
#define AC97_STAC9758_VARIOUS          0x72

 	/* turn on stereo speaker, headphone and line-out */
        AC97_WRITE(ac, AC97_STAC9758_OUTPUT_SELECT, 0x9040);
        /* headphone select and boost */
 	AC97_WRITE(ac, AC97_STAC9758_IO_MISC, 0x2102);
        /* enable mic */
	AC97_WRITE(ac, AC97_STAC9758_INPUT_SELECT, 0x0203);
        /* enable stereo mic */
	AC97_WRITE(ac, AC97_STAC9758_VARIOUS, 0x0001);
    }

#if 0
    /* some experiments, seem to have no effect with a SB16PCI card */
    AC97_WRITE(ac, AC97_MASTERTONE, 0x0f0f);
    AC97_WRITE(ac, AC97_3DCTL,      0x0f0f);
    AC97_WRITE(ac, AC97_GENPURPOSE, 0x7000);
#endif

#if 0
#warning register dump
    {
	int reg;
	char line[100], tmp[8];

	printf("AC97 register dump\n");

	for (reg = 0; reg < 0x80; reg += 0x2) {
	    if (reg % 16 == 0)
		sprintf(line, "%02x:", reg);
	    AC97_READ(ac, reg, &codecval);
	    sprintf(tmp, " %04x", codecval);
	    strcat(line, tmp);
	    switch (reg % 16) {
	    case 6:
		strcat(line, " -");
		break;
	    case 14:
		printf("%s\n", line);
		break;
	    }
	}
    }
#endif

    return 1;
}


/*
 * Handle audio driver "set_config" callbacks from the audio
 * framework.  The audio driver typically just dispatches the
 * "set_config" call to this codec specific implementation.
 */
static int
ac97_set_config(struct audio_codec *base, 
		int command, int dir, int arg1, int arg2)
{
    struct ac97_codec *ac = (struct ac97_codec *)base;
    int config_vol = 0;

    switch (command) {
    case AD_SET_GAIN:
	if (dir == AUDIO_PLAY)
	    ac->audio_info.play.gain = arg1;
	else
	    ac->audio_info.record.gain = arg1;
	config_vol = dir;
	break;
    case AD_SET_GAIN_BAL:
	if (dir == AUDIO_PLAY)
	    ac->audio_info.play.balance = arg1;
	else
	    ac->audio_info.record.balance = arg1;
	config_vol = dir;
	break;
    case AD_SET_PORT:
	if (dir == AUDIO_PLAY)
	    ac->audio_info.play.port = arg1;
	else
	    ac->audio_info.record.port = arg1;
	config_vol = dir;
	break;
    case AD_SET_MONITOR_GAIN:
	ac->audio_info.monitor_gain = arg1;
	config_vol = AUDIO_RECORD;
	break;
    case AD_OUTPUT_MUTE:
	ac->audio_info.output_muted = arg1;
	config_vol = AUDIO_PLAY;
	break;
    case AD_MONO_MIC:
	break;
    case AD_MIC_BOOST:
	break;
    case AD_BASS_BOOST:
	break;
    case AD_MID_BOOST:
	break;
    case AD_TREBLE_BOOST:
	break;
    case AD_LOUDNESS:
	break;
    case AD_SET_DIAG_MODE:
	break;
    }
    
    if (config_vol & AUDIO_PLAY) {
	ac97_configure_playback_vol(ac, &ac->audio_info);
    } else if (config_vol & AUDIO_RECORD) {
	ac97_configure_record_vol(ac, &ac->audio_info);
    }

    return 0;
}


static struct audio_codec_ops ac97_codec_ops = {
    ac97_delete,
    ac97_reset,
    ac97_set_config
};


/*
 * Initialize and return a "codec object" for an AC'97. 
 */
audio_codec_handle_t
ac97_codec_init(audiohdl_t handle,
		struct ac97_codec_ops *ac97_ops, void *ac97_data)
{
    struct ac97_codec *ac;
    dev_info_t *dip = audio_hlp_get_dip(handle);

    ac = kmem_alloc(sizeof(struct ac97_codec), KM_NOSLEEP);
    ac->base.ops = &ac97_codec_ops;
    ac->audio_handle = handle;
    ac->ac97_ops = ac97_ops;
    ac->ac97_data = ac97_data;

    ac->ac97_mic_20db_boost =
	ddi_getprop(DDI_DEV_T_ANY, dip,
		    DDI_PROP_DONTPASS | DDI_PROP_CANSLEEP | DDI_PROP_NOTPROM,
		    "mic20db", 1);

    ac->ac97_pcmvol =
	ddi_getprop(DDI_DEV_T_ANY, dip,
		    DDI_PROP_DONTPASS | DDI_PROP_CANSLEEP | DDI_PROP_NOTPROM,
		    "pcmvol", 8) & 0x1f;

    ac->ac97_eapd =
	ddi_getprop(DDI_DEV_T_ANY, dip,
		    DDI_PROP_DONTPASS | DDI_PROP_CANSLEEP | DDI_PROP_NOTPROM,
		    "eapd", -1);

    return (audio_codec_handle_t)ac;
}



